/*****************************************************************************
*
* File              : TWI_master_slave.c
*							 01.05.09, TAB = 3
* AppNote           : AVR311 / 315
*
*
* nderungen        : Enthlt die Fuktionalit eines Slave und eines Masters.
*							 Luft im Regelfall als (vereinfachter) Slave unabhngig von main():
*							 Wird der Slave schreibend adressiert, dann ist das erste Byte nach der
*							 TWI-Adresse ein Kommando:
*							 Es legt fest, welches Array zum Empfang der nachfolgenden Daten benutzt wird.
*							 Dazu wird lediglich ein Pointer auf das Empfangsarray gesetzt.
*							 Der Index wird auf 0 gesetzt, wenn TWSR = Adressed as ...
*							 Wird der Slave lesend adressiert, dann wird ebenfalls der Index auf 0 gesetzt
*                     und es werden die Daten desjenigen Array gesetzt, auf das der Pointer zeigt.
*
*							 Um ein bestimmtes Array zu beschreiben:
*
*							 TWI_Adresse + w, adr fr gewnschtes Array, Daten ......

*							 Um ein bestimmtes Array auszulesen:
*
*							 TWI_Adresse + w, adr fr gewnschtes Array, TWI_Adress + r, Daten lesen
*
*							 Um Daten aktiv zu versenden, wird in den Master-Modus geschaltet.
*							 Ein Multi-Master-Betrieb ist nicht implementiert, da nur zum Zeitpunkt der
*							 Programmierung mit Schaltdaten ein zweiter Master am Bus aktiv ist.

*                     TWI_slave_mode entscheidet, ob im slave-Modus oder im Master-Modus gearbeitet wird:
*                     in der TWI_ISR werden unterschiedliche switch(TWSR) abgearbeitet, die jeweils ihr
*                     eigenes 'default' habe
*
****************************************************************************/

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include "dcf_77.h"
#include "TWI_master_slave.h"
#include "global.h"


extern void DCF_save_alarm(void);
extern void DCF_delete_alarm(void);
extern void DCF_delete_eeprom(void);

extern volatile uint8_t error;

static uint8_t TWI_MA_buf[MA_BUFFER_SIZE];   // Nach Arbitrierungsfehler bleiben die Sendedaten unverndert erhalten
															// und knnen mit mit ein TWI_MA_Start_Transceiver() neu gesendet werden

static uint8_t TWI_msgSize  = 0;           	// Number of bytes to be transmitted.


// ----------------------------------------------------------------------------
// Grundlegende Initailisierung fr Master und Slave Modus
void TWI_Initialise(void)
{
register uint8_t j;

#ifdef TWI_ADR_CHANGE
//	j = eeprom_read_byte(0);							// passt beim m8 nicht mehr in den Speicher
	while(EECR & (1<<EEWE));							// der manuelle Weg dagegen schon
	EEARH = 0;												// liest die Adresse an Eeprom-Adresse
	EEARL = EEPROM_ADR_LOW;
	EECR |= (1<<EERE);
	j = EEDR;
	EEARL = 0;												// Adresse wieder auf 0 setzen
	if ((j == 255) || (!(j))) j = TWI_SLA_ADR;	//	Adresse aus Eeprom lesen, aber
	j &= 0xFE;												// wenn 255 oder 0, dann den default-Wert bernehmen
	TWAR = j;
#else
	TWAR = TWI_SLA_ADR;
#endif

   TWBR = TWI_TWBR;                             // Set bit rate register (Baudrate). Defined in header file.
// TWSR = TWI_TWPS;                             // Not used. Driver presumes prescaler to be 00.
//	TWDR = 0xFF;                                 // Default content = SDA released.
	TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|      // Prepare to ACK next time the Slave is addressed.
          (0<<TWWC);
	TWI_slave_mode = TRUE;
}

// ----------------------------------------------------------------------------
// Wartet bis der Master seine bertragung beendet hat (erkennbar daran, dass TWIE disabled)
// und schaltet dann wieder zurck in den Slave-Modus
void TWI_SLA_restart(void)
{

while (TWCR & 1<<TWIE);									// warten bis der Master fertig ist
TWI_slave_mode = TRUE;									// Flag fr Slave-Modus setzen
																// TWI-Modul wartet als Slave
TWDR = 0xFF;                                 	// Default content = SDA released.
TWCR = (1<<TWEN)|                            	// TWI Interface enabled.
      (1<<TWIE)|(1<<TWINT)|                 		// Enable TWI Interupt and clear the flag.
      (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|      		// Prepare to ACK next time the Slave is addressed.
      (0<<TWWC);
}


/****************************************************************************
Call this function to send a prepared message. The first byte must contain the slave address and the
read/write bit. Consecutive bytes contain the data to be sent, or empty locations for data to be read
from the slave. Also include how many bytes that should be sent/read including the address byte.
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation,
then initialize the next operation and return.
****************************************************************************/
void TWI_MA_Start_Transceiver_With_Data( uint8_t *msg, uint8_t msgSize )
{
   uint8_t temp;

// while ( TWI_Transceiver_Busy() );           	// Wait until TWI is ready for next transmission.

	TWI_slave_mode = FALSE;								// Flag fr Master-Modus

   TWI_msgSize = msgSize;                      	// Number of data to transmit.
   TWI_MA_buf[0]  = msg[0];                    	// Store slave address with R/W setting.
   if (!( msg[0] & 0x01))						     	// If it is a write operation, then also copy data.
   {
   for ( temp = 1; temp < msgSize; temp++ )
      TWI_MA_buf[ temp ] = msg[ temp ];
   }

   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);
}


/****************************************************************************
This function is the Interrupt Service Routine (ISR), and called when the TWI interrupt is triggered;
that is whenever a TWI event has occurred. This function should not be called directly from the main
application.
****************************************************************************/

ISR(TWI_vect)
{
	static uint8_t TWI_SLA_cmd;						// hier wird im Slavemodus das erste Datenbyte als cmd abgelegt
	static uint8_t *ptr = &time[CLK][0];			// Pointer auf die aktuelle Uhrzeit
	static uint8_t idx;
	static uint8_t first_byte;
	static uint8_t change_adr = FALSE;
if (TWI_slave_mode)
	{
	switch (TWSR)
	{
// TWI Slave Transmitter status codes				// ###############################################

// der Master ist im TX-Modus, der Slave im RX-Modus
// Eigene Slave-Adresse + read empfangen und mit ACK quittiert
   case TWI_STX_ADR_ACK:            				// Own SLA+R has been received; ACK has been returned
      idx = 0;                       				// Set buffer pointer to first data location

// Der Master ist im read-Modus (Slave ist im tx-modus)
// der Slave hat ein Byte gesendet, der Master mit ACK quittiert
   case TWI_STX_DATA_ACK:           				// Data byte in TWDR has been transmitted; ACK has been received
																// Es werden solange Daten gesendet, bis der Master mit NACK beendet
																// damit kann der Pointer aus dem Array hinauslaufen.
//		if (idx < 10) TWDR = ptr[idx++];				// man darf ber die Grenzen eines Arrays hinaus lesen
		TWDR = ptr[idx++];								// - auf eigene Verantwortung

      TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (1<<TWIE)|(1<<TWINT)|              	// Enable TWI Interupt and clear the flag to send byte
            (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   	// Send ACK
            (0<<TWWC);                         	//
      break;

// TWI Slave Receiver status codes					// ###############################################

// der Master ist im TX-Modus, der Slave im RX-Modus
// eigene Slave-Adresse + write empfangen und mit ACK quittiert
   case TWI_SRX_ADR_ACK: 								// Own SLA+W has been received ACK has been returned

		idx = 0;      										// Index fr die nachfolgenden Daten auf 0 stellen
		first_byte = TRUE;								// Flag fr das erste Byte setzen
																// Reset the TWI Interupt to wait for a new event.
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Expect ACK on this transmission
             (0<<TWWC);                         //
      break;

// Master im write Modus (Slave im read Modus), Byte empfangen und mit ACK quittiert
   case TWI_SRX_ADR_DATA_ACK:       				// Previously addressed with own SLA+W; data has been received;
																// ACK has been returned
		if (first_byte)									// das erste Byte
			{
			first_byte = FALSE;							// Flag lschen

			TWI_SLA_cmd = TWDR;							// das erste Datenbyte ist das Kommando
																// es stellt den Pointer auf ein Datenarray


// die Kommandos fr Schaltdefinitionen werden vereinbart als:
// von 0 bis (ALARMDEFS - 1)
// um den Index des zugehrigen Arrays zu ermitteln, muss die Anzahl der vorangestellten
// Zeit-Arrays addiert werden
			if (TWI_SLA_cmd < ALARMDEFS)
				{
				TWI_SLA_cmd += CLK_ARRAYS;
				ptr = &time[TWI_SLA_cmd][0]; 			// hier werden Schaltzeiten an den Pointer gebunden
				}

// wenn die Zeit-Arrays an den Pointer gebunden werden sollen
																// sie werden adressiert mit 0xF0 + Array_index
																// -> siehe  global.h
																// das wichtigste Array: die aktuelle Uhrzeit: 0xF0
			else if ((TWI_SLA_cmd > 0xEF) && (TWI_SLA_cmd < (0xF0 + CLK_ARRAYS)))
				{
				TWI_SLA_cmd -= 0xF0;						// die Zeitarray beginnen an Index 0
				ptr = &time[TWI_SLA_cmd][0]; 			// hier werden die Zeit-Arrays an den Pointer gebunden
				error |= ERR_TWI;							// Fehlerflag setzen, damit die Uhr in der folgenden MInute
				}												// die Zeit manuell incrementiert

// Kommandos
			else if (TWI_SLA_cmd == 0xFA) 			// Schaltzeiten im Eeprom speichern
				{
				ptr = TWI_MA_buf;							// damit nicht versehentlich andere Arrays berschrieben werden
				DCF_save_alarm();
				}

			else if (TWI_SLA_cmd == 0xFB)				// TWI_Adresse ndern, es muss noch auf die Adresse gewartet werden
				{												// die folgt im nchsten Byte
				ptr = TWI_MA_buf;							// damit nicht versehentlich andere Arrays berschrieben werden
				change_adr = TRUE;						// als Flag beim Empfang des 2. Bytes
				}

			else if (TWI_SLA_cmd == 0xFC)				// Alarmdefinitionen aus dem Speicher entfernen
				{
				ptr = TWI_MA_buf;
				DCF_delete_alarm();
				}

			else if (TWI_SLA_cmd == 0xFD)				// Alarmdefinitionen aus dem Eeprom entfernen
				{
				ptr = TWI_MA_buf;
				DCF_delete_eeprom();
				}

			else ptr = TWI_MA_buf;
			}


		else if (change_adr) 							// Sonderfall, die TWI-Adresse soll gendert werden
			{
			change_adr = FALSE;

			EEARH = 0;										// die TWI_Adresse steht am Anfang des Eeproms
			EEARL = EEPROM_ADR_LOW;
			EEDR = (TWDR & 0xFE);

			cli();
			EECR |= (1<<EEMWE);
			EECR |= (1<<EEWE);
			sei();

			TWI_Initialise();								// die genderte Adresse einlesen
			}

		else
			{
			if (idx < 8) ptr[idx++] = TWDR;			// alle Arrays auf 8 Byte + Adresse begrenzen
			}

		TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Send ACK after next reception
             (0<<TWWC);                         //
      break;


// case TWI_STX_DATA_NACK:          				// Data byte in TWDR has been transmitted; NACK has been received.
																// I.e. this could be the end of the transmission.
//	case TWI_STX_ADR_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master;
																// own SLA+R has been received; ACK has been returned
// case TWI_STX_DATA_ACK_LAST_BYTE: 				// Last data byte in TWDR has been transmitted (TWEA = 0);
																// ACK has been received
// case TWI_SRX_ADR_DATA_NACK:      				// Previously addressed with own SLA+W; data has been received;
																// NOT ACK has been returned
// case TWI_SRX_GEN_DATA_NACK:      				// Previously addressed with general call;
																// data has been received; NOT ACK has been returned
// case TWI_SRX_GEN_ACK:            				// General call address has been received; ACK has been returned

// case TWI_SRX_GEN_DATA_ACK:       				// Previously addressed with general call;
																// data has been received; ACK has been returned
// case TWI_SRX_ADR_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master;
																// own SLA+W has been received; ACK has been returned
// case TWI_SRX_GEN_ACK_M_ARB_LOST: 				// Arbitration lost in SLA+R/W as Master;
																// General call address has been received; ACK has been returned
// case TWI_SRX_STOP_RESTART:       				// A STOP condition or repeated START condition has been received
																// while still addressed as Slave
//	case TWI_NO_STATE 									// No relevant state information available; TWINT = 0
// case TWI_BUS_ERROR:         						// Bus error due to an illegal START or STOP condition

// im Slave-Modus laufen alle anderen Flle ber default:
   default:

		TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|              // Enable TWI Interupt and clear the flag to send byte
             (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|   // Send ACK after next reception
             (0<<TWWC);
		break;
	}
	}

else															// TWI-Master-Mode
	{
#ifdef DEBUG
	TWI_MA_buf[9] = TWSR;								// den TWI-Status im letzten Byte sichern
#endif
	switch (TWSR)
	{
// MASTER ########################################################################################

	// General TWI Master status codes				// ###############################################
// ein TWI-(RE)-Start wurde gesendet, nun muss die Adresse folgen
	case TWI_START:              						// START has been transmitted
   case TWI_REP_START:          						// Repeated START has been transmitted
      idx = 0;			           						// Set buffer pointer to the start location

	// TWI Master Transmitter status codes
// die Adresse + W wurde gesendet und vom Slave mit ACK quittiert
   case TWI_MTX_ADR_ACK:        						// SLA+W has been transmitted and ACK received

// eine Datenbyte wurde gesendet und vom Slave mit ACK quittiert
   case TWI_MTX_DATA_ACK:       						// Data byte has been transmitted and ACK received
// wenn noch weitere Bytes abgeholt werden sollen ...
      if (idx < TWI_msgSize)
			{
			TWDR =TWI_MA_buf[idx++];					// Adresse bzw. Daten nach TWDR schreiben
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to send byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|	// Send NACK after reception
               (0<<TWWC);
			}

// alle Bytes gelesen, nun ein Stop senden
// den Rest besorgt 'default'
      else                                      // Send STOP after last byte
			{
//			TWI_statusReg.LastTransOK = TRUE;      // Set status bits to completed successfully.
			TWCR = (1<<TWEN)|                      // TWI Interface enabled
             (0<<TWIE)|(1<<TWINT)|            // Disable TWI Interrupt and clear the flag
             (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|  // Initiate a STOP condition.
             (0<<TWWC);
			}

			break;

// ein Datenbyte wurde empfangen und vom Master mit ACK quittiert
	// TWI Master Receiver status codes				// ###############################################
   case TWI_MRX_DATA_ACK:                       // Data byte has been received and ACK transmitted
      TWI_MA_buf[idx++] = TWDR;						// gelesene Daten in den Puffer schreiben

// die Adresse + R wurde gesendet und vom Slave mit ACK quittiert
   case TWI_MRX_ADR_ACK:                        // SLA+R has been transmitted and ACK received

// es folgen noch weiter Bytes
      if (idx < (TWI_msgSize-1) )        			// Detect the last byte to NACK it.
			{
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send ACK after reception
               (0<<TWWC);
			}
// das folgende zu lesende Byte ist das letzte, danach wird kein ACK (also ein NACK) gesendet
      else                                      // Send NACK after next reception
			{
			TWCR =(1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send NACK after reception
               (0<<TWWC);
			}
      break;

// das letzte Byte wurde empfangen und vom Master mit NACK quittiert
// nun muss der Master ein Stop senden
   case TWI_MRX_DATA_NACK:                      // Data byte has been received and NACK transmitted
      TWI_MA_buf[idx] = TWDR;

      TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (0<<TWIE)|(1<<TWINT)|              	// Disable TWI Interrupt and clear the flag
            (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   	// Initiate a STOP condition.
            (0<<TWWC);
		break;
/*
	// General TWI Master status codes				// ###############################################
   case TWI_ARB_LOST:                           // Arbitration lost
      TWCR =(1<<TWEN)|                         	// TWI Interface enabled
            (1<<TWIE)|(1<<TWINT)|              	// Enable TWI Interupt and clear the flag
            (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|   	// Initiate a (RE)START condition.
            (0<<TWWC);
      break;
*/
// case TWI_MTX_ADR_NACK:                      	// SLA+W has been transmitted and NACK received
// case TWI_MRX_ADR_NACK:                      	// SLA+R has been transmitted and NACK received
// case TWI_MTX_DATA_NACK:                     	// Data byte has been transmitted and NACK received
// case TWI_NO_STATE;                         	// No relevant state information available; TWINT = 0
// case TWI_BUS_ERROR:                        	// Bus error due to an illegal START or STOP condition

   default:

                                                // Reset TWI Interface
      TWCR =(1<<TWEN)|                         	// Enable TWI-interface and release TWI pins
            (0<<TWIE)|(1<<TWINT)|              	// Disable Interupt   - nderung: clear Interrupt Flag
            (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   	// No Signal requests - nderung: Stop Condition senden
            (0<<TWWC);

		break;
		}
	}
}

// -- [ eof ] -----------------------------------------------------------------